#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys, os, re, dbus, subprocess
import gobject, glib
from dbus.mainloop.glib import DBusGMainLoop

# constants
DbusMonitorCmd = "dbus-monitor"
GHelpDbusTarget = "interface=org.gtk.vfs.MountTracker,member=lookupMount"

# DBus signal emitted whenever removable partitions are mounted
DBusRemovableSig = {
	"signal":    "MountAdded",
	"interface": "org.gtk.Private.RemoteVolumeMonitor",
	"path":      "/org/gtk/Private/RemoteVolumeMonitor",
}

# DBus signal emitted whenever video displays are (dis)connected
DBusVideoConnectedSig = {
	"signal":    "OutputConnected",
	"interface": "org.doudoulinux.video",
	"path":      "/org/doudoulinux/video",
}
DBusVideoDisconnectedSig = {
	"signal":    "OutputDisconnected",
	"interface": "org.doudoulinux.video",
	"path":      "/org/doudoulinux/video",
}
DBusVideoNumberSig = {
	"signal":    "NumberActive",
	"interface": "org.doudoulinux.video",
	"path":      "/org/doudoulinux/video",
}
DBusHDMIConnectedSig = {
	"signal":    "HDMIConnected",
	"interface": "org.doudoulinux.video",
	"path":      "/org/doudoulinux/video",
}

#########################
#########################
class DBusMsgMonitor(object):
	"""
	Monitors DBus messages using dbus-monitor.
	This is useful to catch private messages.
	"""
	#########################
	def __init__(self, DBusTarget, Callback):
		"""
		Creates a DBus monitor instance watching given messages and calling a
		callback function to process the received data.
		
		The callback receives data lines of the full DBus message.
		"""
		self.callback = Callback
		self.data = []
		
		# create a pipe running dbus-monitor
		self.command = "%s %s" % (DbusMonitorCmd, DBusTarget)
		self.process = subprocess.Popen(self.command, shell=True, stdout=subprocess.PIPE)
		gobject.io_add_watch(self.process.stdout, glib.IO_IN, self.watch)

	#########################
	def watch(self, FileHandler, Data):
		"""
		Generic callback receiving DBus data into a line buffer.
		"""
		Line = FileHandler.readline()
		if len(Line):
			self.data.append(Line)
			if self.send2callback():
				self.data = []
		
		return True

	#########################
	def send2callback(self):
		"""
		Processes DBus data lines in order to call the callback when a message
		is completed. Returns False if the message cannot be processed yet
		(not enough data) or True if the messages has processed (data buffer
		can be cleared).
		"""
		Buffer = []
		Enclosing = {}
		for Line in self.data:
			if not len(Line.strip()):
				continue
			
			# message start
			if Line[:11] == "method call":
				Buffer.append(Line)
				continue
			if not len(Buffer):
				# dummy data (dbus signal?)
				continue
			
			# register structure opening/closing
			for Delimiter in ("{}", "[]"):
				# track data structure opening
				if Delimiter[0] in Line:
					if Delimiter[0] not in Enclosing:
						Enclosing[Delimiter[0]] = 1
					else:
						Enclosing[Delimiter[0]] += 1
			
				# track data structure closing
				if Delimiter[1] in Line and Delimiter[0] in Enclosing:
					Enclosing[Delimiter[0]] -= 1
				if Delimiter[0] in Enclosing and Enclosing[Delimiter[0]] == 0:
					Enclosing.pop(Delimiter[0])
			
			# track end of data
			Buffer.append(Line)
			if not len(Enclosing):
				self.callback(Buffer)
				return True
		
		return False

#########################
#########################
def DBusMonitor(SignalDict, Callback, monitor="session"):
	"""
	Registers a DBus signal receiver.
	
	monitor = "session" or "system"
	"""
	if monitor == "session":
		monitored_bus = dbus.SessionBus()
	elif monitor == "system":
		monitored_bus = dbus.SystemBus()
	
	monitored_bus.add_signal_receiver(
		signal_name     = SignalDict["signal"],
		dbus_interface  = SignalDict["interface"],
		path            = SignalDict["path"],
		bus_name        = None,
		handler_function= Callback,
	)

#########################
#########################
def start_pcmanfm(Sender, Mount_id, Data):
	"""
	Starts PcManfm whenever a removable media is detected.
	"""
	MountPoint = Data[4][7:]
	print "[dbus-monitor.py] New drive mounted at %s" % MountPoint
	
	if not os.path.isdir(MountPoint):
		print "ERROR: cannot access '%s'" % MountPoint
		return
	
	if subprocess.call("pcmanfm '%s' &" % MountPoint, shell=True):
		print "ERROR: cannot start PcManfm on directory '%s'" % MountPoint

#########################
#########################
def video_connected(CardIds, Status):
	"""
	Activates video output.
	"""
	Name = CardIds["name"]
	Connected = Status["connected"]
	Enabled = Status["enabled"]
	DPMS = Status["dpms"]
	print "[dbus-monitor.py] New video connected: %s (%s, %s, %s)" % (Name, Connected, Enabled, DPMS)
	
	# call xrandr to activate output
	if DPMS == "Off":
		print "[dbus-monitor.py] Activating video output", Name
		if subprocess.call("xrandr --output %s --auto" % Name, shell=True):
			print "ERROR: cannot activate video output '%s'" % Name

#########################
#########################
def video_disconnected(CardIds, Status):
	"""
	De-activates video output.
	"""
	Name = CardIds["name"]
	Connected = Status["connected"]
	Enabled = Status["enabled"]
	DPMS = Status["dpms"]
	print "[dbus-monitor.py] Video output disconnected: %s (%s, %s, %s)" % (Name, Connected, Enabled, DPMS)
	
	# call xrandr to deactivate output
	if DPMS == "On":
		print "[dbus-monitor.py] Deactivating video output", Name
		if subprocess.call("xrandr --output %s --off" % Name, shell=True):
			print "ERROR: cannot deactivate video output '%s'" % Name

#########################
#########################
def video_selector(Data):
	"""
	Starts lxrandr when two video displays are active.
	"""
	Number = int(Data["number_active"])
	List = Data["list"]
	print "[dbus-monitor.py] %d active video output(s): %s" % (Number, List)
	
	# call lxrandr to let user choose screen arrangement
	if Number > 1:
		if subprocess.call("lxrandr &", shell=True):
			print "ERROR: cannot start Lxrandr"

#########################
#########################
def hdmi_connected(CardIds, Status):
	"""
	Shows the sound output dialog whenever HDMI is connected/disconnected.
	"""
	Name = CardIds["name"]
	Connected = Status["connected"]
	Enabled = Status["enabled"]
	DPMS = Status["dpms"]
	print "[dbus-monitor.py] HDMI connection modified: %s (%s, %s, %s)" % (Name, Connected, Enabled, DPMS)
	
	# call set-alsa-output to let user decide and test sound output
	print "[dbus-monitor.py] Showing set-alsa-output", Name
	if subprocess.call("gksu set-alsa-output &", shell=True):
		print "ERROR: cannot launch 'set-alsa-output'"

#########################
#########################
def check_dbus_ghelp(LineList):
	"""
	Checks the DBus messages are actually calling for gnome-help then
	calls gnome-help with the URI extracted from DBus data.
	"""
	# look for ghelp in the request
	Data = "".join(LineList)
	if 'array of bytes "ghelp"' not in Data:
		return

	# extract URI
	URI = re.findall('"([^"]+)"', LineList[2])
	if not len(URI):
		print "ERROR: cannot find URI in line '%s'" % LineList[2]
		return
	URI = URI[0]

	# call gnome-help
	print "[dbus-monitor.py] Calling ghelp on URI", URI
	if re.findall('^/[^/]+$', URI):
		# case of URI = /gedit
		subprocess.call("gnome-help ghelp:" + URI.lstrip("/") + " &", shell=True)
	else:
		# case of URI = /usr/share/gnome/help/gnome-terminal/fr/gnome-terminal.xml
		subprocess.call("gnome-help ghelp://" + URI + " &", shell=True)


#########################
#########################
#########################
#########################
DBusGMainLoop(set_as_default=True)

# monitor removable disks by registering a DBus signal receiver
DBusMonitor(DBusRemovableSig, start_pcmanfm)

# monitor video displays (dis)connected by registering a DBus signal receiver
DBusMonitor(DBusVideoConnectedSig, video_connected, monitor="system")
DBusMonitor(DBusVideoDisconnectedSig, video_disconnected, monitor="system")
DBusMonitor(DBusVideoNumberSig, video_selector, monitor="system")

# monitor HDMI screens (dis)connected by registering a DBus signal receiver
DBusMonitor(DBusHDMIConnectedSig, hdmi_connected, monitor="system")

# monitor ghelp calls by watching the output of dbus-monitor (this is private message)
CallMonitor = DBusMsgMonitor(GHelpDbusTarget, check_dbus_ghelp)

# enter the event loop
loop = gobject.MainLoop()
loop.run()
