#!/bin/sh
#
# DoudouLinux udev/dbus script library
#
# Requires dbus, alsa-utils.
#
# Author: JM. Philippe <jean-michel.philippe@doudoulinux.org>
#################################
# This file is part of DoudouLinux.
#
# DoudouLinux is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DoudouLinux is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with DoudouLinux.  If not, see <http://www.gnu.org/licenses/>.

# contants
# NB: $ACTION, $DEVPATH, $DEVNAME are global vars set by udev
GLOBALSTATUS=/var/tmp/connected-video-outputs
DBUS_DEST=org.doudoulinux.video
DBUS_PATH=/org/doudoulinux/video
DBUS_IF_ON=org.doudoulinux.video.OutputConnected
DBUS_IF_OFF=org.doudoulinux.video.OutputDisconnected
DBUS_IF_NBR=org.doudoulinux.video.NumberActive
DBUS_IF_HDMI=org.doudoulinux.video.HDMIConnected

# source doudoulinux-sound function library
FUNCLIB=/usr/lib/doudoulinux-sound/function-lib.sh
if ! [ -f $FUNCLIB ]; then
	echo "/etc/udev/doudoulinux-video: cannot find '$FUNCLIB'"
	exit 1
fi
. $FUNCLIB

#######################
video_output_name() {
	# computes the video output name seen by xrandr
	#
	# video_output_name SYSPATH OUTPUT
	SYSPATH=$1
	OUTPUT=$2
	
	# remove '-' and isolated letters in name
	# example: LVDS-1 -> LVDS1, HDMI-A-1 -> HDMI1
	CARDNAME=$(basename $SYSPATH)
	CARDNAME=$(echo $OUTPUT | sed -r 's/'$CARDNAME'-([^-]+).*([^-]+)$/\1\2/')
	
	# special case of TV output
	if [ "${CARDNAME#SVIDEO}" != "$CARDNAME" ]; then
		# this is a TV composite output
		echo TV${CARDNAME#SVIDEO}
	fi
	
	# return name if any
	echo "$CARDNAME"
}

#######################
create_statusfile() {
	# creates our global status file containing the list of active video outputs
	#
	# create_statusfile SYSPATH
	SYSPATH=$1
	
	# put in this file the video outputs connected *and* enabled
	# we assume only connected and enabled outputs were connected before
	# we don't care of DPMS Off since we list available screens here
	CONNECTED=
	for STATUSFILE in $(find $SYSPATH -name status); do
		ENABLEFILE=$(dirname $STATUSFILE)/enabled
		if [ -f $ENABLEFILE ]; then
			if ( [ "$(cat $STATUSFILE)" = "connected" ] && [ "$(cat $ENABLEFILE)" = "enabled" ] ); then
				OUTPUT=$(basename "$(dirname $STATUSFILE)")
				CONNECTED="$CONNECTED $OUTPUT"
			fi
		fi
	done
	echo "$CONNECTED" > $GLOBALSTATUS
	chmod 644 $GLOBALSTATUS
}

#######################
output_param() {
	# get a video output parameter, TYPE can be:
	#   - enabled (enabled/disabled)
	#   - status (connected/disconnected)
	#   - dpms (On/Off)
	# "xrandr --off …" puts DPMS status to Off, but output is still enabled
	#
	# output_param TYPE SYSPATH OUTPUT
	TYPE=$1
	SYSPATH=$2
	OUTPUT=$3
	
	QUERYFILE=$SYSPATH/$OUTPUT/$TYPE
	if [ -f $QUERYFILE ]; then
		cat $QUERYFILE
	fi
}

#######################
alsa_hdmi_output() {
	# returns the ALSA HDMI output associated with video card number
	# we assume video card0 will output on HDMI card0
	#
	# alsa_hdmi_output XRANDNAME
	XRANDNAME=$1
	
	# search for an HDMI audio device
	if [ "${XRANDNAME#HDMI}" != "$XRANDNAME" ]; then
		NBR=$(( ${XRANDNAME#HDMI} + 1 ))
		AUDIOCARD=$(aplay -l | grep '^card .*HDMI' | head -n $NBR | tail -n 1)
		if [ -n "$AUDIOCARD" ]; then
			# return audio output number
			echo $AUDIOCARD | sed -r 's/^card +([0-9]+).+/\1/'
			return 0
		fi
	fi
	return 1
}

#######################
dbus_video_notification() {
	# sends a dbus message as 2 dictionaries:
	# one for output ids, the other one for output status
	#
	# dbus_video_notification SYSPATH OUTPUT XRANDNAME DBUS_IFACE
	SYSPATH=$1
	OUTPUT=$2
	XRANDNAME=$3
	DBUS_IFACE=$4
	
	CONNECTED=$(output_param "status" $SYSPATH $OUTPUT)
	ENABLED=$(output_param "enabled" $SYSPATH $OUTPUT)
	DPMS=$(output_param "dpms" $SYSPATH $OUTPUT)
	dbus-send --system --type=signal $DBUS_PATH $DBUS_IFACE \
		dict:string:string:path,"$SYSPATH",name,"$XRANDNAME",output,"$OUTPUT" \
		dict:string:string:connected,"$CONNECTED",enabled,"$ENABLED",dpms,"$DPMS"
}

#######################
scan_connected() {
	# scan connected outputs and send a dbus message
	# if they were not previously connected
	#
	# scan_connected SYSPATH
	SYSPATH=$1
	
	CONNECTEDLIST=
	for STATUSFILE in $(find $SYSPATH -name status); do
		OUTPUT=$(basename "$(dirname $STATUSFILE)")
		
		# we can't trust TV (S-video) output status
		XRANDNAME=$(video_output_name $SYSPATH $OUTPUT)
		if [ "${XRANDNAME#TV}" != "$XRANDNAME" ]; then
			CONNECTED=$(output_param "status" $SYSPATH $OUTPUT)
			DPMS=$(output_param "dpms" $SYSPATH $OUTPUT)
			if ( [ "$CONNECTED" = "disconnected" ] && [ "$DPMS" = "On" ] ); then
				# it may happen that TV output is active while not being connected?
				dbus_video_notification $SYSPATH $OUTPUT $XRANDNAME $DBUS_IF_OFF
			fi
			continue
		fi
		
		# output newly connected
		if [ "$(cat $STATUSFILE)" = "connected" ]; then
			CONNECTEDLIST="$CONNECTEDLIST $OUTPUT"
			
			# we don't do anything if the output is already listed in our global status file
			if ! grep "$OUTPUT" $GLOBALSTATUS >/dev/null; then
				# retrieve the newest status of the output
				# then send a dbus system signal
				dbus_video_notification $SYSPATH $OUTPUT $XRANDNAME $DBUS_IF_ON
			fi
			
			# if this is an HDMI output, there may be an audio output associated
			if [ "${XRANDNAME#HDMI}" != "$XRANDNAME" ]; then
				# search for an HDMI audio device
				CARDNBR=$(alsa_hdmi_output $XRANDNAME)
				if [ $? -eq 0 ]; then
					# change audio output to the detected audio HDMI output
					set_audiocard --keep $CARDID
					restart_timidity
				fi
				
				# send a dbus notification
				dbus_video_notification $SYSPATH $OUTPUT $XRANDNAME $DBUS_IF_HDMI
			fi
		fi
	done
	
	# update global list of outputs
	echo "$CONNECTEDLIST" > $GLOBALSTATUS
}

#######################
find_disconnected() {
	# finds disconnected outputs and send a dbus message
	# if they were previously connected
	#
	# find_disconnected SYSPATH
	SYSPATH=$1
	
	# first of all re-activate screens that are DPMS Off to avoid an
	# X server crash if only the disconnected screen was really active
	# NB: the disconnected screen can be DPMS On or Off
	for OUTPUT in $(cat $GLOBALSTATUS); do
		# we can't trust TV (S-video) output status
		XRANDNAME=$(video_output_name $SYSPATH $OUTPUT)
		if [ "${XRANDNAME#TV}" != "$XRANDNAME" ]; then
			continue
		fi
		
		# try to re-activate sleeping screens
		# NB: GDM retarts the X server as soon as the running output is disconnected,
		#     this is a bit useless then.
		CONNECTED=$(output_param "status" $SYSPATH $OUTPUT)
		if [ "$CONNECTED" = "connected" ]; then
			DPMS=$(output_param "dpms" $SYSPATH $OUTPUT)
			if [ "$DPMS" = "Off" ]; then
				dbus_video_notification $SYSPATH $OUTPUT $XRANDNAME $DBUS_IF_ON
			fi
		fi
	done
	
	# we can now stop video output for the removed screen(s)
	for OUTPUT in $(cat $GLOBALSTATUS); do
		# we can't trust TV (S-video) output status
		XRANDNAME=$(video_output_name $SYSPATH $OUTPUT)
		if [ "${XRANDNAME#TV}" != "$XRANDNAME" ]; then
			continue
		fi
		
		# output newly disconnected
		CONNECTED=$(output_param "status" $SYSPATH $OUTPUT)
		if [ "$CONNECTED" = "disconnected" ]; then
			dbus_video_notification $SYSPATH $OUTPUT $XRANDNAME $DBUS_IF_OFF
			
			# if this is an HDMI output, there may be an audio output associated
			if [ "${XRANDNAME#HDMI}" != "$XRANDNAME" ]; then
				# search for an HDMI audio device
				CARDNBR=$(alsa_hdmi_output $XRANDNAME)
				if [ $? -eq 0 ]; then
					# change audio outputs to its initial value
					init_audiocard
					restart_timidity
				fi
				
				# send a dbus notification
				dbus_video_notification $SYSPATH $OUTPUT $XRANDNAME $DBUS_IF_HDMI
			fi
		fi
	done
}

#######################
tell_active() {
	# sends a dbus message telling how many screen are now active
	#
	# tell_active SYSPATH
	SYSPATH=$1
	
	# send a dbus message telling how many active screen there are
	LIST=$(cat $GLOBALSTATUS | sed -r 's/^[ ]+//')
	NBR=$(echo $LIST | wc -w)
	dbus-send --system --type=signal $DBUS_PATH $DBUS_IF_NBR \
		dict:string:string:path,"$SYSPATH",number_active,$NBR,list,"$LIST"
}
